//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <d3dx8.h>
#include "Sound.h"
#include "Vertex.h"
#include "Aqua.h"


Aqua::Aqua()
{
    Device = NULL;
    VertexBuffer = NULL;
    MaxPrimitives = 0;

    DrawWidth = GRID_WIDTH;
    DrawHeight = GRID_HEIGHT;

    WaterDensity = 50;
    WaterHeight = 256;

    Viewer = NULL;

    for(long i=0;i<MAX_RAIN;i++)
    {
        Rain[i] = RainDrop();
    }
}

Aqua::~Aqua()
{
    SAFE_RELEASE(VertexBuffer);
    SAFE_RELEASE(SplashVertexBuffer);
    SAFE_RELEASE(SmokeTexture);
    SAFE_RELEASE(SplashTexture);
}

void Aqua::Initialize(LPDIRECT3DDEVICE8 dev,D3DXVECTOR3 *v)
{
    Device = dev;

    UINT num = ((MAX_RAIN>>3)*2);
    Device->CreateVertexBuffer(num*sizeof(ParticleVertex),D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC,Vertex_FVF,D3DPOOL_DEFAULT,&VertexBuffer);

    num = MAX_RAIN;
    Device->CreateVertexBuffer(num*sizeof(ParticleVertex),D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC,Vertex_FVF,D3DPOOL_DEFAULT,&SplashVertexBuffer);

    //find out some things about the capabilities
    D3DCAPS8 Caps;
    Device->GetDeviceCaps(&Caps);
    MaxPrimitives = Caps.MaxPrimitiveCount;

    D3DXCreateTextureFromFile(Device,"Textures\\Splash.bmp",&SplashTexture);
    D3DXCreateTextureFromFile(Device,"Textures\\Smoke.bmp",&SmokeTexture);

    //load the sounds
    RainSound = LoadSoundSSF("Sounds\\Rain.ssf");
    SetSoundVolume(RainSound,((float)(100-WaterDensity))/100*(0.1));
    PlaySound(RainSound,TRUE);
    LavaSplash[0] = LoadSoundSSF("Sounds\\LavaSplash.ssf");
    LavaSplash[1] = LoadSoundSSF("Sounds\\LavaSplash2.ssf");
    MakeCopiesOfSound(LavaSplash[0]);
    MakeCopiesOfSound(LavaSplash[1]);
    SetSoundVolume(LavaSplash[0],0.2);
    SetSoundVolume(LavaSplash[1],0.2);
    Viewer = v;
}

void Aqua::Update(float MoveSpeed,Terra *Earth)
{
    static draw_odd=TRUE;
    draw_odd = !draw_odd;

    SetSoundVolume(RainSound,((float)(100-WaterDensity))/100*(0.1));

    for(long i=draw_odd;i<MAX_RAIN;i+=2)
    {
        if(Rain[i].Active)
        {
            if(Rain[i].Pos.y<0)
            {
                Rain[i].Active = FALSE;
                CheckCollision(Earth,i);
            }

            Rain[i].Pos += Rain[i].Vel*MoveSpeed;
            Rain[i].Vel.y -= MoveSpeed*(Rain[i].Length/8);
        }

        SplashUpdate(MoveSpeed,draw_odd ? i : i-1);
        SplashUpdate(MoveSpeed,draw_odd ? i+1 : i);
    }
    if(draw_odd)
        RainDropCreate(MoveSpeed,WaterDensity);
}

void Aqua::SplashUpdate(float MoveSpeed,long i)
{
    DWORD temp;
    float tempf=0.0;
    if(Splash[i].Active)
    {
        Splash[i].Pos += Splash[i].Vel;
        Splash[i].Vel *= 0.92;

        if(!Splash[i].Smoke)
        {
            Splash[i].Vel.y -= (MoveSpeed/4);
            if(Splash[i].Pos.y<0)
                Splash[i].Vel.y = -Splash[i].Vel.y;
        }

        temp = GetTickCount()-Splash[i].StartTime;
        if(temp>Splash[i].Life)
            Splash[i].Active = FALSE;
        else
        {
            tempf = (float)temp/(float)Splash[i].Life;
            tempf = 1-tempf;
            Splash[i].Vertex.color = D3DCOLOR_COLORVALUE(tempf,tempf,tempf,tempf);
        }
    }
}

void Aqua::RainDropCreate(float MoveSpeed,long num)
{
    long activated=0;
    for(long i=0;i<MAX_RAIN;i++)
    {
        if(activated >= num)
            break;

        if(!Rain[i].Active)
        {
            Rain[i].Active = TRUE;
            Rain[i].Pos = D3DXVECTOR3(RandomNumber(GRID_SIZE*(DrawWidth>>1)),WaterHeight,RandomNumber(GRID_SIZE*(DrawHeight>>1)));
            Rain[i].Vel = D3DXVECTOR3(0,-(MoveSpeed/2),0);
            Rain[i].Length = RandomNumber(24)+8;
            Rain[i].LeadVertex.color = D3DCOLOR_COLORVALUE(0,0,RandomNumber(0.5)+0.5,1);
            Rain[i].TailVertex.color = Rain[i].LeadVertex.color;
            activated++;
        }
    }
}

void Aqua::SplashCreate(D3DXVECTOR3 pos,D3DXVECTOR3 vel,WORD life,long num,BOOL smoke)
{
    long activated=0;
    float atten=0.0;
    for(long i=0;i<MAX_RAIN;i++)
    {
        if(activated>=num)
            break;

        if(!Splash[i].Active)
        {
            Splash[i].Pos = pos;
            Splash[i].Vel = vel;
            Splash[i].Life = life;
            Splash[i].Active = TRUE;
            Splash[i].StartTime = GetTickCount();
            Splash[i].Smoke = smoke;
            if(Viewer!=NULL)
                atten = (DISTANCE(*Viewer,Splash[i].Pos)/512);
            if(smoke && (rand()%10==0) && (atten<1.0))
                PlaySound(LavaSplash[rand()%2],FALSE,atten);
            activated++;
        }
    }
}

void Aqua::Draw(void)
{
    Device->SetStreamSource(0,VertexBuffer,sizeof(ParticleVertex));
    Device->SetVertexShader(ParticleVertex_FVF);
    Device->SetTexture(0,NULL);

    Device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    Device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
    Device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

    DWORD num_to_draw=0;
    ParticleVertex *buffer=0;
    VertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);
    for(int q=0;q<4;q++)    //render the 4 quadrants
    {
      for(long i=0;i<MAX_RAIN;i++)
      {
       if(num_to_draw<(MAX_RAIN>>3))
       {
        if(Rain[i].Active)
        {
            Rain[i].Refresh(q,DrawWidth,DrawHeight);
            memcpy(buffer,&Rain[i].LeadVertex,sizeof(ParticleVertex));
            buffer++;
            memcpy(buffer,&Rain[i].TailVertex,sizeof(ParticleVertex));
            buffer++;

            num_to_draw++;
        }
       }
       else
       {
        VertexBuffer->Unlock();
        Device->DrawPrimitive(D3DPT_LINELIST,0,num_to_draw);
        VertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);
        num_to_draw=0;
       }
      }
      if(num_to_draw>0)
      {
        VertexBuffer->Unlock();
        Device->DrawPrimitive(D3DPT_LINELIST,0,num_to_draw);
      }

    }

    Device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    SplashDraw();
}

void Aqua::SplashDraw(void)
{
    Device->SetStreamSource(0,SplashVertexBuffer,sizeof(ParticleVertex));
    Device->SetVertexShader(ParticleVertex_FVF);
    Device->SetTexture(0,SplashTexture);

    Device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    Device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
    Device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
    Device->SetRenderState(D3DRS_POINTSPRITEENABLE, TRUE);
    Device->SetRenderState(D3DRS_POINTSCALEENABLE, TRUE);
    Device->SetRenderState(D3DRS_POINTSIZE_MIN, FToDW(0.00) );
    Device->SetRenderState(D3DRS_POINTSIZE_MAX, FToDW(64.00) );
    Device->SetRenderState(D3DRS_POINTSIZE, FToDW(4.0) );
    Device->SetRenderState(D3DRS_POINTSCALE_A, FToDW(0.00) );
    Device->SetRenderState(D3DRS_POINTSCALE_B, FToDW(0.00) );
    Device->SetRenderState(D3DRS_POINTSCALE_C, FToDW(8.00) );
    Device->SetRenderState(D3DRS_ZWRITEENABLE, FALSE );

    ///////////////////////////////////////////////
    // first pass: all the splash water textures //
    ///////////////////////////////////////////////
    DWORD num_to_draw=0;
    ParticleVertex *buffer=0;
    SplashVertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);
    for(int i=0;i<MAX_RAIN;i++)
    {
      if(num_to_draw<(MAX_RAIN/10))
      {
        if(Splash[i].Active && !Splash[i].Smoke)
        {
            Splash[i].Refresh();
            memcpy(buffer,&Splash[i].Vertex,sizeof(ParticleVertex));
            buffer++;
            num_to_draw++;
        }
      }
      else
      {
        SplashVertexBuffer->Unlock();
        Device->DrawPrimitive(D3DPT_POINTLIST,0,num_to_draw);
        SplashVertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);
        num_to_draw=0;
      }
    }
    if(num_to_draw>0)
    {
        SplashVertexBuffer->Unlock();
        Device->DrawPrimitive(D3DPT_POINTLIST,0,num_to_draw);
    }
    /////////////////////////////////////////
    // second pass: all the smoke textures //
    /////////////////////////////////////////
    Device->SetTexture(0,SmokeTexture);
    Device->SetRenderState(D3DRS_POINTSIZE, FToDW(16.0) );
    Device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCCOLOR);
    Device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCCOLOR);

    num_to_draw=0;
    SplashVertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);
    for(int i=0;i<MAX_RAIN;i++)
    {
      if(num_to_draw<(MAX_RAIN/10))
      {
        if(Splash[i].Active && Splash[i].Smoke)
        {
            Splash[i].Refresh();
            memcpy(buffer,&Splash[i].Vertex,sizeof(ParticleVertex));
            buffer++;
            num_to_draw++;
        }
      }
      else
      {
        SplashVertexBuffer->Unlock();
        Device->DrawPrimitive(D3DPT_POINTLIST,0,num_to_draw);
        SplashVertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);
        num_to_draw=0;
      }
    }
    if(num_to_draw>0)
    {
        SplashVertexBuffer->Unlock();
        Device->DrawPrimitive(D3DPT_POINTLIST,0,num_to_draw);
    }


    Device->SetRenderState(D3DRS_POINTSCALEENABLE, FALSE);
    Device->SetRenderState(D3DRS_POINTSPRITEENABLE, FALSE);
    Device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    Device->SetRenderState(D3DRS_ZWRITEENABLE, TRUE );
}

void Aqua::CheckCollision(Terra *Earth,long num)
{
    float dist=0.0;
    D3DXVECTOR3 tempvec(0,0,0);
    BOOL earth_collide = 0;
    int x=0,y=0;

    for(int q=0;q<4;q++)
    {
      earth_collide = FALSE;
      tempvec = Rain[num].Pos;

      if(q == 1 || q==3)
        tempvec.z += (GRID_SIZE*(DrawHeight>>1));
      if(q == 2 || q==3)
        tempvec.x += (GRID_SIZE*(DrawWidth>>1));

      y = (int)(tempvec.z/GRID_SIZE);
      x = (int)(tempvec.x/GRID_SIZE);
      if(y>Earth->DrawHeight)
        y = Earth->DrawHeight;
      if(x>Earth->DrawWidth)
        x = Earth->DrawWidth;
      if(y<0)
        y = 0;
      if(x<0)
        x = 0;

      try
      {
       if(!Earth->Grid[(y*GRID_HEIGHT)+x].Active )
       {
        dist = DISTANCE_NOY(Earth->Grid[(y*GRID_HEIGHT)+x].Pos,tempvec);
        if(dist<GRID_SIZE)
        {
            SplashCreate(D3DXVECTOR3(tempvec.x,0,tempvec.z),D3DXVECTOR3(RandomNumber(1)-0.5,3,RandomNumber(1)-0.5),500,1,FALSE);
            earth_collide = TRUE;
        }
       }
       if(!earth_collide)
           SplashCreate(D3DXVECTOR3(tempvec.x,-(GRID_SIZE>>1),tempvec.z),D3DXVECTOR3(0,RandomNumber(0.5)+0.75,0),2000,1,TRUE);
      }
      catch(...)
      {
      }
    }
}

//---------------------------------------------------------------------------
