//---------------------------------------------------------------------------
#ifndef AquaH
#define AquaH

#include "Vertex.h"
#include "Terra.h"
#include "Ignis.h"
#include <d3dx8.h>

#define MAX_RAIN 5000

typedef struct _RainDrop
{
    _RainDrop()
    {
        Pos = D3DXVECTOR3(0,64,0);
        Vel = D3DXVECTOR3(0,-1,0);

        Length = 16.0;

        Active = FALSE;

        LeadVertex = ParticleVertex(D3DXVECTOR3(0,0,0),D3DCOLOR_RGBA(0,0,255,128));
        TailVertex = ParticleVertex(D3DXVECTOR3(0,0,0),D3DCOLOR_RGBA(0,0,255,128));
    }

    void Refresh(int quadrant,int w,int h)
    {
        LeadVertex.x = Pos.x;
        LeadVertex.y = Pos.y;
        LeadVertex.z = Pos.z;

        switch(quadrant)
        {
            default:
            break;

            case 1:
            LeadVertex.z += (GRID_SIZE*(h>>1));
            break;

            case 2:
            LeadVertex.x += (GRID_SIZE*(w>>1));
            break;

            case 3:
            LeadVertex.x += (GRID_SIZE*(w>>1));
            LeadVertex.z += (GRID_SIZE*(h>>1));
            break;
        }

        D3DXVECTOR3 norm;
        D3DXVec3Normalize(&norm,&Vel);
        TailVertex.x = LeadVertex.x - norm.x * Length;
        TailVertex.y = LeadVertex.y - norm.y * Length;
        TailVertex.z = LeadVertex.z - norm.z * Length;
    }

    D3DXVECTOR3 Pos;
    D3DXVECTOR3 Vel;

    float Length;

    BOOL Active;

    ParticleVertex LeadVertex;  //to form a line
    ParticleVertex TailVertex;

}RainDrop;

typedef struct _SplashParticle
{
    _SplashParticle()
    {
        Pos = D3DXVECTOR3(0,0,0);
        Vel = D3DXVECTOR3(0,0,0);

        Life = 500;
        StartTime = GetTickCount();

        Active = FALSE;
        Smoke = rand()%2;

        Vertex = ParticleVertex(D3DXVECTOR3(0,0,0),D3DCOLOR_RGBA(255,255,255,255));
    }

    void Refresh(void)
    {
        Vertex.x = Pos.x;
        Vertex.y = Pos.y;
        Vertex.z = Pos.z;
    }

    D3DXVECTOR3 Pos;
    D3DXVECTOR3 Vel;

    WORD Life;
    long StartTime;

    BOOL Active;
    BOOL Smoke;

    ParticleVertex Vertex;

}SplashParticle;

class Aqua
{
public:

    Aqua();
    ~Aqua();

    void Initialize(LPDIRECT3DDEVICE8 dev,D3DXVECTOR3 *v);

    void Update(float MoveSpeed,Terra *Earth);
    void SplashUpdate(float MoveSpeed,long i);

    void RainDropCreate(float MoveSpeed,long num);
    void SplashCreate(D3DXVECTOR3 pos,D3DXVECTOR3 vel,WORD life,long num,BOOL smoke);

    void Draw(void);
    void SplashDraw(void);

    void CheckCollision(Terra *Earth,long num);

    long RainSound;
    long LavaSplash[2]; //when the rain hits the lava
    D3DXVECTOR3 *Viewer;    //controls sound attenuation

    WORD WaterDensity;
    WORD WaterHeight;

    WORD DrawWidth;
    WORD DrawHeight;
    RainDrop Rain[MAX_RAIN];
    SplashParticle Splash[MAX_RAIN];

    DWORD MaxPrimitives;
    LPDIRECT3DDEVICE8 Device;
    LPDIRECT3DVERTEXBUFFER8 VertexBuffer;
    LPDIRECT3DVERTEXBUFFER8 SplashVertexBuffer;
    LPDIRECT3DTEXTURE8 SmokeTexture;
    LPDIRECT3DTEXTURE8 SplashTexture;
};

//---------------------------------------------------------------------------
#endif
