//---------------------------------------------------------------------------
#include <vcl.h>
#include <d3dx8.h>
#pragma hdrstop

#include "Ignis.h"
#include "Sound.h"
#include <algorithm>
#include <cmath>
#include <time.h>

Ignis::Ignis()
{
    Device = NULL;
    VertexBuffer = NULL;
    IndexBuffer = NULL;
    Texture = NULL;

    Theta = 0.0;

    NumTriangles = 0;

    ParticleSize = 32.0;

    float temp=0.0;
    float tempy=0.0;
    for(int y=0;y<GRID_HEIGHT;y++)
        for(int x=0;x<GRID_WIDTH;x++)
        {
            temp = RandomNumber(1.0);
            tempy = -(GRID_SIZE>>1);
            LavaGrid[(y*GRID_HEIGHT)+x] = Vertex(D3DXVECTOR3((x*GRID_SIZE)+(GRID_SIZE>>1),tempy,(y*GRID_SIZE)+(GRID_SIZE>>1)),D3DCOLOR_COLORVALUE((temp<0.25) ? temp+0.75 : 1.0,(temp>=0.25) ? temp+0.25 : 0.0,0.0,1.0),(x%2),(y%2));
            VelocityGrid[(y*GRID_HEIGHT)+x] = VelocityStruct();
            VelocityGrid[(y*GRID_HEIGHT)+x].OriginalY = tempy;

            LavaWaveArray[(y*GRID_HEIGHT)+x] = 0.0;
        }

    for(int i=0;i<MAX_PARTICLES;i++)
        Particles[i] = LavaParticle();
}

Ignis::~Ignis()
{
    SAFE_RELEASE(VertexBuffer);
    SAFE_RELEASE(IndexBuffer);
    SAFE_RELEASE(Texture);
}

void Ignis::Initialize(LPDIRECT3DDEVICE8 Dev)
{
    Device = Dev;

    UINT num = (GRID_HEIGHT*GRID_WIDTH);
    Device->CreateVertexBuffer(num*sizeof(Vertex),D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC,Vertex_FVF,D3DPOOL_DEFAULT,&VertexBuffer);

    //find out some things about the capabilities
    D3DCAPS8 Caps;
    Device->GetDeviceCaps(&Caps);
    MaxPrimitives = Caps.MaxPrimitiveCount;

    //setup the index buffer
    Device->CreateIndexBuffer(((GRID_WIDTH-1)*(GRID_HEIGHT-1)*6)*sizeof(WORD),D3DUSAGE_WRITEONLY,D3DFMT_INDEX16,D3DPOOL_DEFAULT,&IndexBuffer);

    WORD Indices[((GRID_WIDTH-1)*(GRID_HEIGHT-1)*6)];
    WORD curr_index=0;

    NumTriangles = 0; //total number of triangles to draw (take off the last row)
    for(int y=0;y<(GRID_HEIGHT-1);y++)
    {
        for(int x=0;x<(GRID_WIDTH-1);x++) //fill in the indices 2 triangles at a time
        {
            Indices[curr_index] = ( ((y)*GRID_HEIGHT) + (x) );
            Indices[curr_index+1] = ( ((y+1)*GRID_HEIGHT) + (x) );
            Indices[curr_index+2] = ( ((y+1)*GRID_HEIGHT) + (x+1) );

            curr_index+=3;
            NumTriangles++;

            Indices[curr_index] = ( ((y+1)*GRID_HEIGHT) + (x+1) );
            Indices[curr_index+1] = ( ((y)*GRID_HEIGHT) + (x+1) );
            Indices[curr_index+2] = ( ((y)*GRID_HEIGHT) + (x) );

            curr_index+=3;
            NumTriangles++;
        }
    }

    void *buffer;
    IndexBuffer->Lock(0,0,(BYTE**)&buffer,0);
    memcpy(buffer,Indices,sizeof(Indices));
    IndexBuffer->Unlock();

    //setup the texture
    D3DXCreateTextureFromFileEx(Device,"Textures\\Lava.bmp",0,0,D3DX_DEFAULT,0,D3DFMT_UNKNOWN,D3DPOOL_MANAGED,D3DX_DEFAULT,D3DX_DEFAULT,0,NULL,NULL,&Texture);

    /////////////////////////////////
    //setup the particle variables //
    /////////////////////////////////

    Device->CreateVertexBuffer(PARTICLES_PER_EXPLOSION*sizeof(ParticleVertex),D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC,Vertex_FVF,D3DPOOL_DEFAULT,&ParticleVertexBuffer);
    D3DXCreateTextureFromFileEx(Device,"Textures\\LavaParticle.bmp",0,0,D3DX_DEFAULT,0,D3DFMT_UNKNOWN,D3DPOOL_MANAGED,D3DX_DEFAULT,D3DX_DEFAULT,0,NULL,NULL,&ParticleTexture);

    //load the sound
    LavaGurgleSound = LoadSoundSSF("Sounds\\LavaGurgle.ssf");
    PlaySound(LavaGurgleSound,TRUE);
    EruptionSound = LoadSoundSSF("Sounds\\Eruption.ssf");
    BigEruptionSound = LoadSoundSSF("Sounds\\Eruption_Big.ssf");

    MakeCopiesOfSound(EruptionSound);
    MakeCopiesOfSound(BigEruptionSound);
}

void Ignis::Update(float MoveSpeed,D3DXVECTOR3 viewer)
{
    Theta += 0.01;
    float tempx=Theta;
    float tempy=0.0;
    float tempz=Theta;
    float tempavg=0.0;
    long i=0;

    static long clock_last=0;
    static long clock_diff=0;
    static long clock_last_2=0;
    static long clock_diff_2=0;

    clock_diff -= (clock_last-(clock()));
    clock_diff_2 -= (clock_last_2-(clock()));

    for(long y=0;y<GRID_HEIGHT;y++)
    {
      for(long x=0;x<GRID_WIDTH;x++)
      {
        i = (long)((y*GRID_HEIGHT)+x);

        if(!VelocityGrid[i].Active) //make the lava sinuous
        {
            LavaGrid[((y*GRID_HEIGHT)+x)].x = ((x*GRID_SIZE)+(GRID_SIZE>>1))+sin(tempx)*(GRID_SIZE>>2);
            LavaGrid[((y*GRID_HEIGHT)+x)].z = ((y*GRID_SIZE)+(GRID_SIZE>>1))+cos(tempz)*(GRID_SIZE>>2);

            LavaGrid[((y*GRID_HEIGHT)+x)].u += 0.0014;
            LavaGrid[((y*GRID_HEIGHT)+x)].v += 0.0014;

            if(clock_diff>20) 
            {
              if(y<(GRID_HEIGHT-1) && y>0 && x<(GRID_WIDTH-1) && x>0 )
              {
                tempy = (GRID_SIZE>>1);
                tempavg = ((LavaWaveArray[(((y+1)*GRID_HEIGHT)+(x))]-tempy) +
                            (LavaWaveArray[(((y)*GRID_HEIGHT)+(x+1))]-tempy) +
                            (LavaWaveArray[(((y-1)*GRID_HEIGHT)+(x))]-tempy) +
                            (LavaWaveArray[(((y)*GRID_HEIGHT)+(x-1))]-tempy))/4;

                LavaGrid[((y*GRID_HEIGHT)+x)].y = tempavg+tempy;
              }
            }
        }
        else
        {
            if(VelocityGrid[i].Bounces<4)
            {
                LavaGrid[i].y += VelocityGrid[i].Vel.y*MoveSpeed;
                VelocityGrid[i].Vel.y -= MoveSpeed/2;

                if(LavaGrid[i].y<(VelocityGrid[i].OriginalY))
                {
                    VelocityGrid[i].Vel.y = -VelocityGrid[i].Vel.y;
                    VelocityGrid[i].Bounces++;
                    VelocityGrid[i].Vel.y *= 0.5; //slow it down (very bouncy bounces)

                    if(VelocityGrid[i].Bounces==4)
                    {
                        LavaGrid[i].y = VelocityGrid[i].OriginalY;
                        VelocityGrid[i].Vel.y = 0.0;
                    }
                }
                if(x%2==0 && y%2==0 && clock_diff_2>100) //make the plumes react to the wind
                    InitParticle(LavaGrid[i].ToVector(),D3DXVECTOR3(0,MoveSpeed,0),1000);
            }
            else
            {
                VelocityGrid[i].Vel *= 0.00; //slow it down more, for now
                VelocityGrid[i].Active = FALSE;
            }
        }
        tempx +=((D3DX_PI*2)/(GRID_WIDTH/8));
      }
      tempz+=((D3DX_PI*2)/(GRID_HEIGHT/8));
    }
    if(clock_diff>20)
        clock_diff=0;
    clock_last = clock();
    
    if(clock_diff_2>100)
        clock_diff_2=0;
    clock_last_2 = clock();

    for(long y=0;y<GRID_HEIGHT;y++) //save the wave height array
    {
      for(long x=0;x<GRID_WIDTH;x++)
      {
        i = ((y*GRID_HEIGHT)+x);
        LavaWaveArray[i] = LavaGrid[i].y;
      }
    }

    //create random bits of fire on the lava's surface
    InitParticle(D3DXVECTOR3(RandomNumber(GRID_SIZE*GRID_WIDTH),0,RandomNumber(GRID_SIZE*GRID_HEIGHT)),D3DXVECTOR3(RandomNumber(8)-4,0,RandomNumber(8)-4),RandomNumber(1000)+500);
    InitParticle(D3DXVECTOR3(RandomNumber(GRID_SIZE*GRID_WIDTH),0,RandomNumber(GRID_SIZE*GRID_HEIGHT)),D3DXVECTOR3(RandomNumber(8)-4,0,RandomNumber(8)-4),RandomNumber(1000)+500);

    ParticleUpdate(MoveSpeed);

    SetSoundVolume(LavaGurgleSound,(viewer.y/512)*0.5);
}

void Ignis::ParticleUpdate(float MoveSpeed)
{
    DWORD diff=0;
    DWORD temp = 0;
    for(long i=0;i<MAX_PARTICLES;i++)
    {
        if(!Particles[i].Active)
            continue;

        Particles[i].Pos += Particles[i].Vel*MoveSpeed;
        Particles[i].Vel.y -= (MoveSpeed/2);

        if(Particles[i].Pos.y<0)
        {
            Particles[i].Vel.y = -Particles[i].Vel.y;
            Particles[i].Vel *= 0.75;
        }

        diff = (GetTickCount()-Particles[i].StartTime);
        if(diff > Particles[i].Life)
            Particles[i].Active = FALSE;
        else
        {
            if(Particles[i].Life!=0)
            {
                temp = (UCHAR)((1-((float)((float)diff)/Particles[i].Life))*255);
                Particles[i].Vertex.color = D3DCOLOR_RGBA(temp,temp,temp,temp);
            }
            else
                Particles[i].Vertex.color = D3DCOLOR_RGBA(255,255,255,0);
        }
    }
}

void Ignis::Draw(float yrot)
{
    Device->SetIndices(IndexBuffer,0);
    Device->SetStreamSource(0,VertexBuffer,sizeof(Vertex));
    Device->SetVertexShader(Vertex_FVF);
    Device->SetTexture(0,Texture);

    void *buffer=NULL;
    VertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);
    memcpy(buffer,LavaGrid,sizeof(LavaGrid));
    VertexBuffer->Unlock();

    DWORD i=(GRID_HEIGHT*GRID_WIDTH);
    DWORD num_to_draw=0; //per loop
    DWORD num_drawn=0;

    for(;;) //send as many triangles as possible (about 2048), or send them as much at a time as possible
    {
        if(num_drawn<NumTriangles)
        {
            num_to_draw = (MaxPrimitives < (NumTriangles-num_drawn)) ? MaxPrimitives : (NumTriangles-num_drawn);
            Device->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,i,(num_drawn*3),num_to_draw);
            num_drawn += num_to_draw;
        }
        else
            break;
    }

    ParticleDraw(yrot);
}

void Ignis::ParticleDraw(float CameraYRot)
{
    Device->SetStreamSource(0,ParticleVertexBuffer,sizeof(ParticleVertex));
    Device->SetVertexShader(ParticleVertex_FVF);
    Device->SetTexture(0,ParticleTexture);

    Device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    Device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
    Device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
    Device->SetRenderState(D3DRS_POINTSPRITEENABLE, TRUE);
    Device->SetRenderState(D3DRS_POINTSCALEENABLE, TRUE);
    Device->SetRenderState(D3DRS_POINTSIZE_MIN, FToDW(0.00) );
    Device->SetRenderState(D3DRS_POINTSIZE_MAX, FToDW(64.00) );
    Device->SetRenderState(D3DRS_POINTSIZE, FToDW(32.0) );
    Device->SetRenderState(D3DRS_POINTSCALE_A, FToDW(0.00) );
    Device->SetRenderState(D3DRS_POINTSCALE_B, FToDW(0.00) );
    Device->SetRenderState(D3DRS_POINTSCALE_C, FToDW(8.00) );
    Device->SetRenderState(D3DRS_ZWRITEENABLE, FALSE );


    DWORD num_to_draw=0;
    ParticleVertex *buffer=0;
    ParticleVertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);
    for(int i=0;i<MAX_PARTICLES;i++)
    {
      if(num_to_draw<PARTICLES_PER_EXPLOSION)
      {

        if(Particles[i].Active)
        {
            Particles[i].RefreshVertices();
            memcpy(buffer,&Particles[i].Vertex,sizeof(ParticleVertex));
            buffer++;
            num_to_draw++;
        }
      }
      else
      {
        ParticleVertexBuffer->Unlock();
        Device->DrawPrimitive(D3DPT_POINTLIST,0,num_to_draw);
        ParticleVertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);
        num_to_draw=0;
      }
    }
    if(num_to_draw>0)
    {
        ParticleVertexBuffer->Unlock();
        Device->DrawPrimitive(D3DPT_POINTLIST,0,num_to_draw);
    }

    Device->SetRenderState(D3DRS_POINTSCALEENABLE, FALSE);
    Device->SetRenderState(D3DRS_POINTSPRITEENABLE, FALSE);
    Device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    Device->SetRenderState(D3DRS_ZWRITEENABLE, TRUE );
}

void Ignis::Explode(D3DXVECTOR3 pos,long radius,float magnitude,Terra *Ground)
{
    float distance;
    float mag_times_dist=0.0;

    for(long y=((long)pos.z-radius);y<((long)pos.z+radius);y++)
    {
      if(y<0 || y>=GRID_HEIGHT)
        continue;

      for(long x=((long)pos.x-radius);x<((long)pos.x+radius);x++)
      {
        if(x<0 || x>=GRID_WIDTH)
            continue;
        if(VelocityGrid[(y*GRID_HEIGHT)+x].Active)
            continue;

        distance = DISTANCE_NOY(D3DXVECTOR3((float)pos.x*GRID_SIZE,pos.y,(float)pos.z*GRID_SIZE),LavaGrid[(y*GRID_HEIGHT)+x].ToVector());
        if(distance > (float)(radius*GRID_SIZE))
            continue;

        mag_times_dist = magnitude*(1-(distance/(radius*GRID_SIZE)));
        VelocityGrid[(y*GRID_HEIGHT)+x].Vel.y = mag_times_dist;
        VelocityGrid[(y*GRID_HEIGHT)+x].Bounces = 0;
        VelocityGrid[(y*GRID_HEIGHT)+x].Active = TRUE;
      }
    }
    ParticleExplode(D3DXVECTOR3(pos.x*GRID_SIZE,pos.y,pos.z*GRID_SIZE),(radius*GRID_SIZE),magnitude);
    if(Ground!=NULL)
        Ground->Explode(D3DXVECTOR3(pos.x*GRID_SIZE,pos.y,pos.z*GRID_SIZE),(radius*GRID_SIZE),magnitude);

    //make the explosion sound!!
    if(magnitude>20)
        PlaySound(BigEruptionSound,FALSE);
    else
    {
        SetSoundVolume(EruptionSound, ((15-(magnitude-5))/15)*0.2 );
        PlaySound(EruptionSound,FALSE);
    }
}

void Ignis::ParticleExplode(D3DXVECTOR3 pos,float radius, float magnitude)
{
    long num_active=0;
    D3DXVECTOR3 normalized;
    float mag_times_distance=0.0;
    float distance = 0.0;

    for(long i=0;i<MAX_PARTICLES;i++)
    {
        if(num_active>=PARTICLES_PER_EXPLOSION)
            break;

        if(!Particles[i].Active)
        {
            Particles[i].Active = TRUE;
            Particles[i].Pos = D3DXVECTOR3(pos.x+(RandomNumber(radius*2)-radius),0,pos.z+(RandomNumber(radius*2)-radius));

            distance = DISTANCE_NOY(pos,Particles[i].Pos);
            D3DXVec3Normalize(&normalized,&(Particles[i].Pos - pos));

            mag_times_distance = magnitude*(1-(distance/radius));
            Particles[i].Vel = normalized*mag_times_distance;

            Particles[i].StartTime = GetTickCount();
            Particles[i].Life = 1000+(DWORD)(RandomNumber(1000)-500);

            num_active++;
        }
    }

}

void Ignis::InitParticle(D3DXVECTOR3 pos, D3DXVECTOR3 vel, DWORD life)
{
    for(long i=0;i<MAX_PARTICLES;i++)
    {
        if(!Particles[i].Active)
        {
            Particles[i].Active = TRUE;
            Particles[i].Pos = pos;
            Particles[i].Vel = vel;
            Particles[i].Life = life;
            Particles[i].StartTime = GetTickCount();
            break;
        }
    }
}

//---------------------------------------------------------------------------
