//---------------------------------------------------------------------------
#ifndef IgnisH
#define IgnisH

#include <d3dx8.h>
#include "Terra.h"  //to keep track of the grid variables
#include "Vertex.h"

//---------------------------------------------------------------------------

#define MAX_PARTICLES 10000
#define PARTICLES_PER_EXPLOSION 200
#define SET_VERTEX_POS(a,b) a.x=b.x; a.y=b.y; a.z=b.z;

typedef struct _LavaParticle
{
    _LavaParticle()
    {
        Pos = D3DXVECTOR3(0,0,0);
        Vel = D3DXVECTOR3(0,0,0);
        Active = FALSE;
        Life = 0;
        StartTime = GetTickCount();

        Vertex = ParticleVertex(D3DXVECTOR3(0,0,0),D3DCOLOR_RGBA(255,255,255,0));
    }

    void RefreshVertices(void)
    {
        Vertex.x = Pos.x;
        Vertex.y = Pos.y;
        Vertex.z = Pos.z;
    }

    D3DXVECTOR3 Pos; //position
    D3DXVECTOR3 Vel; //velocity

    DWORD Life;       //how long it lives (in seconds)
    DWORD StartTime;  //when it was created

    BOOL Active;
    ParticleVertex Vertex;

}LavaParticle;

typedef struct _VelocityStruct
{
    _VelocityStruct()
    {
        Vel = D3DXVECTOR3(0,0,0);
        Bounces = 10;
        Active = FALSE;
        OriginalY = 0.0;
    }

    D3DXVECTOR3 Vel;
    float OriginalY;
    WORD Bounces;
    BOOL Active;

}VelocityStruct;

class Ignis //Fire and Lava
{
public:
    Ignis();
    ~Ignis();

    void Initialize(LPDIRECT3DDEVICE8 Dev);
    void Update(float MoveSpeed,D3DXVECTOR3 viewer);
    void Draw(float yrot);

    void Explode(D3DXVECTOR3 pos,long radius,float magnitude,Terra *Ground=NULL);

    float LavaWaveArray[GRID_WIDTH*GRID_HEIGHT]; //used to average the lava down (a seeping effect)
    Vertex LavaGrid [GRID_WIDTH*GRID_HEIGHT];   //a grid of points that control the lava flow
    VelocityStruct VelocityGrid[GRID_WIDTH*GRID_HEIGHT];    //an array that describes the above grid's velocity, activity, and bounces

    //particle variables/functions
    void ParticleUpdate(float MoveSpeed);
    void ParticleDraw(float CameraYRot);
    void ParticleExplode(D3DXVECTOR3 pos,float radius,float magnitude);
    void InitParticle(D3DXVECTOR3 pos,D3DXVECTOR3 vel,DWORD life); 

    float ParticleSize;
    LavaParticle Particles[MAX_PARTICLES];  //used for extra fiery effects
    LPDIRECT3DVERTEXBUFFER8 ParticleVertexBuffer;
    LPDIRECT3DTEXTURE8 ParticleTexture;

    DWORD NumTriangles;
    DWORD MaxPrimitives;    //maximum number of primitives this device can handle at once

    float Theta;

    long LavaGurgleSound;
    long EruptionSound;
    long BigEruptionSound;

    LPDIRECT3DDEVICE8 Device;
    LPDIRECT3DVERTEXBUFFER8 VertexBuffer;
    LPDIRECT3DINDEXBUFFER8 IndexBuffer;
    LPDIRECT3DTEXTURE8 Texture;
};

#endif
