//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <dsound.h>
#include <stdio.h>
#include "Sound.h"

LPDIRECTSOUND8 DirectSound=NULL;
SoundStruct Sounds[MAX_SOUNDS];
BOOL PrimaryFormatSet=FALSE;

void InitializeDirectSound(HWND window)
{
    static int init=1;
    if(init)
    {
        DirectSoundCreate8(NULL,&DirectSound,NULL);
        DirectSound->SetCooperativeLevel(window,DSSCL_PRIORITY);
        for(long i=0;i<MAX_SOUNDS;i++)
        {
            Sounds[i] = SoundStruct();
        }
    }
    init = 0;
}

void DestroyDirectSound(void)
{
    StopAllSounds();
    ReleaseAllSounds();
    DirectSound->Release();
    DirectSound = NULL;
}

void PlaySound(long index,BOOL loop,float volume)
{
    if(Sounds[index].Loaded && Sounds[index].Buffer!=NULL)
    {
        if(!Sounds[index].bCopies)
            Sounds[index].Buffer->Play(0,0,loop ? DSBPLAY_LOOPING : 0);
        else
        {
            DWORD status;
            for(int i=0;i<MAX_COPIES;i++)
            {
                Sounds[index].Copies[i]->GetStatus(&status);
                if(!(status & DSBSTATUS_PLAYING))
                {
                    if(volume!=-999.999)
                        Sounds[index].Copies[i]->SetVolume((long)(volume*-10000));
                    Sounds[index].Copies[i]->Play(0,0,loop ? DSBPLAY_LOOPING : 0);
                    break;
                }
            }
        }
    }
}

void StopSound(long index)
{
    if(Sounds[index].Loaded && Sounds[index].Buffer!=NULL)
    {
        if(!Sounds[index].bCopies)
            Sounds[index].Buffer->Stop();
        else
        {
            for(int i=0;i<MAX_COPIES;i++)
            {
                Sounds[index].Copies[i]->Stop();
            }
        }
    }
}

void SetSoundVolume(long index,float volume)
{
    if(volume<0)
        volume = -volume;
    if(volume>1)
        volume = 1;

    if(Sounds[index].Loaded && Sounds[index].Buffer!=NULL)
    {
        if(!Sounds[index].bCopies)
            Sounds[index].Buffer->SetVolume((long)(volume*-10000));
        else
        {
            for(int i=0;i<MAX_COPIES;i++)
                Sounds[index].Copies[i]->SetVolume((long)(volume*-10000));
        }
    }
}

void StopAllSounds(void)
{
    for(long i=0;i<MAX_SOUNDS;i++)
    {
        StopSound(i);
    }
}

void ReleaseAllSounds(void)
{
    for(long i=0;i<MAX_SOUNDS;i++)
    {
        if(Sounds[i].Loaded && Sounds[i].Buffer!=NULL)
        {
            Sounds[i].Buffer->Release();
            Sounds[i].Buffer = NULL;
            Sounds[i].Loaded = FALSE;
        }
    }
}

void MakeCopiesOfSound(long index)
{
    if(Sounds[index].Loaded && Sounds[index].Buffer!=NULL)
    {
        Sounds[index].bCopies = TRUE;
        for(int i=0;i<MAX_COPIES;i++)
        {
            if(DirectSound->DuplicateSoundBuffer(Sounds[index].Buffer,&Sounds[index].Copies[i]) != DS_OK)
            {
                Sounds[index].bCopies = FALSE;
                break;
            }
        }
    }
}

/*/////////////////////////////////////////////////////////////////////

    SSF stands for simple sound format and was created by me
    specifically for this contest.

    It consists of a simple header, followed by the raw data.

    I did this becuase VOC files were to low-quality and WAV files
    were to complicated.

*//////////////////////////////////////////////////////////////////////

long LoadSoundSSF(char *filename)
{
    int data_offset,         // offset to data part of file
    	playback_rate,       // playback rate as encoded in file
    	data_length;         // length of data

    ULONG filelength = 0;

    FILE  *f;

    UCHAR *snd_buffer,       // temporary sound buffer to hold voc data
          *audio_ptr_1=NULL, // data ptr to first write buffer
    	  *audio_ptr_2=NULL; // data ptr to second write buffer

    DWORD audio_length_1=0,  // length of first write buffer
    	  audio_length_2=0;  // length of second write buffer

    long  sound_id=-1;

    //find an unloaded sound
    for(sound_id=0;sound_id<=MAX_SOUNDS;sound_id++)
    {
        if(sound_id==MAX_SOUNDS)
            return(-1);
        else
        if(!Sounds[sound_id].Loaded && Sounds[sound_id].Buffer==NULL)
            break;
    }

    if ((f = fopen(filename,"rb"))==NULL)
	    return(0);

    //find the file length
    fseek(f,0,SEEK_END);
    filelength = ftell(f);
    fseek(f,0,SEEK_SET);

    // now read in the header
    SSFHeader h;
    fread(&h,sizeof(h),1,f);

    //now allocate and read in the data buffer
    snd_buffer = (UCHAR *)malloc(h.DataBytes);
    fseek(f,sizeof(h),SEEK_SET);
    for(long i=0;i<h.DataBytes;i++)
        fread(&snd_buffer[i],1,1,f);

    // access all values
    data_offset	  = sizeof(h);
    playback_rate = h.SampleRate;
    data_length   = h.DataBytes;

    // set rate and size in data structure
    Sounds[sound_id].PlaybackRate  = playback_rate;
    Sounds[sound_id].Size  = data_length;

    fclose(f);

    // set up the format data structure
    WAVEFORMATEX pcmwf;
    memset(&pcmwf, 0, sizeof(WAVEFORMATEX));

    pcmwf.wFormatTag	  = WAVE_FORMAT_PCM;
    pcmwf.nChannels		  = 1;
    pcmwf.nSamplesPerSec  = h.SampleRate;
    pcmwf.nBlockAlign	  = (h.BitsPerSample/8);
    pcmwf.nAvgBytesPerSec = pcmwf.nSamplesPerSec * pcmwf.nBlockAlign;
    pcmwf.wBitsPerSample  = h.BitsPerSample;
    pcmwf.cbSize		  = 0;

    // prepare to create sounds buffer
    DSBUFFERDESC dsbd;
    ZeroMemory(&dsbd,sizeof(DSBUFFERDESC));
    dsbd.dwSize			= sizeof(DSBUFFERDESC);
    dsbd.dwFlags		= DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | DSBCAPS_STATIC | DSBCAPS_LOCSOFTWARE;
    dsbd.dwBufferBytes	= data_length;
    dsbd.lpwfxFormat	= &pcmwf;

    // create the sound buffer
    if (DirectSound->CreateSoundBuffer(&dsbd,&Sounds[sound_id].Buffer,NULL)!=DS_OK)
    {
        // release memory
        free(snd_buffer);

        // return error
        return(-1);
    } // end if

    // copy data into sound buffer
    if (Sounds[sound_id].Buffer->Lock(0,
                        data_length,
                        (void **) &audio_ptr_1,
                        &audio_length_1,
                        (void **)&audio_ptr_2,
                        &audio_length_2,
                        DSBLOCK_FROMWRITECURSOR)!=DS_OK)
        return(0);

    // copy first section of circular buffer
    memcpy(audio_ptr_1, snd_buffer, audio_length_1);

    // copy last section of circular buffer
    memcpy(audio_ptr_2, (snd_buffer+audio_length_1),audio_length_2);

    // unlock the buffer
    if (Sounds[sound_id].Buffer->Unlock(audio_ptr_1,
                          audio_length_1,
                          audio_ptr_2,
                          audio_length_2)!=DS_OK)
        return(0);

    // release the temp buffer
    free(snd_buffer);

    Sounds[sound_id].Loaded = TRUE;

    if(!PrimaryFormatSet)
    {
        Sounds[sound_id].Buffer->SetFormat(&pcmwf);
        PrimaryFormatSet = TRUE;
    }

    // return id
    return(sound_id);

} // end LoadSoundSSF

/*////////////////////////////////////////////////////////////

    The following code is originally from the book
    'Windows Game Programming for Dummies' by Andre LaMothe.
    This new functions is based on his code and the information found in
    chapter 13 on DirectSound, but has been edited to fit
    this context and programming style.

////////////////////////////////////////////////////////////*/

long LoadSoundVOC(char *filename)
{
    int data_offset,         // offset to data part of file
    	playback_rate,       // playback rate as encoded in file
    	data_length;         // length of data

    ULONG filelength = 0;

    FILE  *f;

    UCHAR *snd_buffer,       // temporary sound buffer to hold voc data
          *audio_ptr_1=NULL, // data ptr to first write buffer
    	  *audio_ptr_2=NULL; // data ptr to second write buffer

    DWORD audio_length_1=0,  // length of first write buffer
    	  audio_length_2=0;  // length of second write buffer

    long  sound_id=-1;

    //find an unloaded sound
    for(sound_id=0;sound_id<=MAX_SOUNDS;sound_id++)
    {
        if(sound_id==MAX_SOUNDS)
            return(-1);
        else
        if(!Sounds[sound_id].Loaded && Sounds[sound_id].Buffer==NULL)
            break;
    }

    if ((f = fopen(filename,"rb"))==NULL)
	    return(0);

    //find the file length
    fseek(f,0,SEEK_END);
    filelength = ftell(f);
    fseek(f,0,SEEK_SET);

    // allocate a large enough temporary buffer
    snd_buffer = (UCHAR *)malloc(filelength);

    // now read in the data
    fread(snd_buffer,filelength,1,f);

    // access all values and decode VOC encoding for data fields
    data_offset	  = snd_buffer[20];
    playback_rate = (-1000000/(snd_buffer[data_offset+4]-256));
    data_length   = ((*(int *)(snd_buffer+data_offset)) >> 8);

    // set rate and size in data structure
    Sounds[sound_id].PlaybackRate  = playback_rate;
    Sounds[sound_id].Size  = data_length;

    fclose(f);

    // set up the format data structure
    WAVEFORMATEX pcmwf;
    memset(&pcmwf, 0, sizeof(WAVEFORMATEX));

    pcmwf.wFormatTag	  = WAVE_FORMAT_PCM;
    pcmwf.nChannels		  = 1;
    pcmwf.nSamplesPerSec  = 22222;
    pcmwf.nBlockAlign	  = 1;
    pcmwf.nAvgBytesPerSec = pcmwf.nSamplesPerSec * pcmwf.nBlockAlign;
    pcmwf.wBitsPerSample  = 8;
    pcmwf.cbSize		  = 0;

    // prepare to create sounds buffer
    DSBUFFERDESC dsbd;
    ZeroMemory(&dsbd,sizeof(DSBUFFERDESC));
    dsbd.dwSize			= sizeof(DSBUFFERDESC);
    dsbd.dwFlags		= DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | DSBCAPS_STATIC | DSBCAPS_LOCSOFTWARE;
    dsbd.dwBufferBytes	= data_length-NVB_SIZE;
    dsbd.lpwfxFormat	= &pcmwf;

    // create the sound buffer
    if (DirectSound->CreateSoundBuffer(&dsbd,&Sounds[sound_id].Buffer,NULL)!=DS_OK)
    {
        // release memory
        free(snd_buffer);

        // return error
        return(-1);
    } // end if

    // copy data into sound buffer
    if (Sounds[sound_id].Buffer->Lock(0,
                        data_length-NVB_SIZE,
                        (void **) &audio_ptr_1,
                        &audio_length_1,
                        (void **)&audio_ptr_2,
                        &audio_length_2,
                        DSBLOCK_FROMWRITECURSOR)!=DS_OK)
        return(0);

    // copy first section of circular buffer
    memcpy(audio_ptr_1, snd_buffer+data_offset+NVB_SIZE, audio_length_1);

    // copy last section of circular buffer
    memcpy(audio_ptr_2, (snd_buffer+data_offset+NVB_SIZE+audio_length_1),audio_length_2);

    // unlock the buffer
    if (Sounds[sound_id].Buffer->Unlock(audio_ptr_1,
                          audio_length_1,
                          audio_ptr_2,
                          audio_length_2)!=DS_OK)
        return(0);

    // release the temp buffer
    free(snd_buffer);

    Sounds[sound_id].Loaded = TRUE;

    // return id
    return(sound_id);

} // end LoadSoundVOC

//---------------------------------------------------------------------------
