//---------------------------------------------------------------------------
#include <vcl.h>
#include <d3dx8.h>
#pragma hdrstop

#include "Vertex.h"
#include "Terra.h"
#include "Sound.h"
#include <cstdlib>

//Terra class definitions//

Terra::Terra()
{
    Device = NULL;
    VertexBuffer = NULL;
    Texture = NULL;
    MaxPrimitives = 0;
    Changed = TRUE;

    DrawWidth = GRID_WIDTH;
    DrawHeight = GRID_HEIGHT;

    Viewer = NULL;

    float temp=0.0;
    randomize();
    for(int y=0;y<GRID_HEIGHT;y++)
        for(int x=0;x<GRID_WIDTH;x++)
        {
            Grid[((y)*GRID_HEIGHT)+x] = GridInfo();

            if(x==0 || x==(GRID_WIDTH-1) || y==0 || y==(GRID_HEIGHT-1))
                Grid[((y)*GRID_HEIGHT)+x].Pos = D3DXVECTOR3((x*GRID_SIZE),0,(y*GRID_SIZE));
            else
                Grid[((y)*GRID_HEIGHT)+x].Pos = D3DXVECTOR3((x*GRID_SIZE)+RandomNumber(GRID_SIZE/2),RandomNumber(12)-(GRID_SIZE>>1),(y*GRID_SIZE)+RandomNumber(GRID_SIZE/2));  //add some randomness

            temp = RandomNumber(0.5)+0.5;
            Grid[((y)*GRID_HEIGHT)+x].R = (UCHAR)(temp*255);
            Grid[((y)*GRID_HEIGHT)+x].G = (UCHAR)((RandomNumber(0.25)+0.75)*255);
            Grid[((y)*GRID_HEIGHT)+x].B = (UCHAR)(temp*255);
        }
    for(int y=0;y<GRID_HEIGHT;y++)//with all the if else's this looks real messy...
        for(int x=0;x<GRID_WIDTH;x++)
        {
            if(y==(GRID_HEIGHT-1))
            {
                if(x==(GRID_WIDTH-1))
                {
                    Grid[((y)*GRID_HEIGHT)+x].Init(
                        Grid[((y)*GRID_HEIGHT)+(x)],
                        Grid[((y)*GRID_HEIGHT)+(x)],
                        Grid[((y)*GRID_HEIGHT)+(x)]);
                }
                else
                {
                    Grid[((y)*GRID_HEIGHT)+x].Init(
                        Grid[((y)*GRID_HEIGHT)+(x+1)],
                        Grid[((y)*GRID_HEIGHT)+(x)],
                        Grid[((y)*GRID_HEIGHT)+(x+1)]);
                }
            }
            else
            if(x==(GRID_WIDTH-1))
            {
                Grid[((y)*GRID_HEIGHT)+x].Init(
                        Grid[((y)*GRID_HEIGHT)+(x)],
                        Grid[((y+1)*GRID_HEIGHT)+(x)],
                        Grid[((y+1)*GRID_HEIGHT)+(x)]);
            }
            else
            {
                Grid[((y)*GRID_HEIGHT)+x].Init(
                        Grid[((y)*GRID_HEIGHT)+(x+1)],
                        Grid[((y+1)*GRID_HEIGHT)+(x)],
                        Grid[((y+1)*GRID_HEIGHT)+(x+1)]);
            }
        }
}

Terra::~Terra()
{
    SAFE_RELEASE(VertexBuffer);
    SAFE_RELEASE(IndexBuffer);
    SAFE_RELEASE(Texture);
}

void Terra::Initialize(LPDIRECT3DDEVICE8 Dev,D3DXVECTOR3 *v)
{
    Device = Dev;

    //find out some things about the capabilities
    D3DCAPS8 Caps;
    Device->GetDeviceCaps(&Caps);
    MaxPrimitives = Caps.MaxPrimitiveCount;

    UINT num = 8;
    Device->CreateVertexBuffer(num*(GRID_WIDTH)*sizeof(Vertex),D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC,Vertex_FVF,D3DPOOL_DEFAULT,&VertexBuffer);

    //setup the index buffer
    Device->CreateIndexBuffer(18*(GRID_WIDTH)*sizeof(WORD),D3DUSAGE_WRITEONLY,D3DFMT_INDEX16,D3DPOOL_DEFAULT,&IndexBuffer);

    WORD Indices[] =
    {
      2,3,0,
      3,1,0,
      0,1,4,
      1,3,5,
      3,2,6,
      2,0,7
    };
    BYTE *buffer;
    IndexBuffer->Lock(0,0,(BYTE**)&buffer,0);
    for(long i=0;i<(GRID_WIDTH);i++)
    {
        memcpy(buffer,Indices,sizeof(Indices));
        buffer+=sizeof(Indices);
        for(int i=0;i<18;i++)
            Indices[i] += 8;
    }
    IndexBuffer->Unlock();

    //create the texture
    D3DXCreateTextureFromFileEx(Device,"Textures\\Earth.bmp",0,0,D3DX_DEFAULT,0,D3DFMT_UNKNOWN,D3DPOOL_MANAGED,D3DX_DEFAULT,D3DX_DEFAULT,0,NULL,NULL,&Texture);

    //create the vertex shader
    DWORD VSDesc[] =
    {
        D3DVSD_STREAM(0),
        D3DVSD_REG( D3DVSDE_POSITION, D3DVSDT_FLOAT3),
        D3DVSD_REG( D3DVSDE_DIFFUSE, D3DVSDT_D3DCOLOR),
        D3DVSD_REG( D3DVSDE_TEXCOORD0, D3DVSDT_FLOAT2),
        D3DVSD_END()
    };

    //Load sounds
    RockHits[0] = LoadSoundSSF("Sounds\\RockHit1.ssf");
    RockHits[1] = LoadSoundSSF("Sounds\\RockHit2.ssf");
    RockHits[2] = LoadSoundSSF("Sounds\\RockHit3.ssf");
    RockHits[3] = LoadSoundSSF("Sounds\\RockHit4.ssf");
    RockHits[4] = LoadSoundSSF("Sounds\\RockHit5.ssf");
    for(int i=0;i<5;i++)
        MakeCopiesOfSound(RockHits[i]);
    Viewer = v;
}

void Terra::Update(float MoveSpeed,float RotateSpeed)
{
    BOOL temp=TRUE;
    long i=0;
    float atten=0.0;

    for(long y=0;y<DrawHeight;y++)
    {
      for(long x=0;x<DrawWidth;x++)
      {
        i = ((GRID_HEIGHT*y)+x);

        if(!Grid[i].Active)
            continue;

        temp=FALSE;
        if(Grid[i].Bounces<3)
        {
            if(Grid[i].Pos.y<-4)
            {
                //make a hitting sound//
                if(Viewer != NULL)
                    atten = (DISTANCE(*Viewer,Grid[i].Pos)/1024);
                if(i%5==0 && atten<1.0)
                    PlaySound(RockHits[rand()%5],FALSE,atten);

                Grid[i].Velocity.y = -Grid[i].Velocity.y;
                Changed = TRUE;
                Grid[i].Bounces++;
                Grid[i].Velocity.y *= 0.2; //slow it down (very shallow bounces)

                Grid[i].YawV *= 0.75;
                Grid[i].PitchV *= 0.75;
                Grid[i].RollV *= 0.75;

                if(Grid[i].Bounces==3)
                {
                    Grid[i].Pos.y = 0.0;
                    Grid[i].Velocity.y = 0.0;
                }
            }
            if(Grid[i].Pos.x>(GRID_SIZE*GRID_WIDTH) || Grid[i].Pos.x<0)
                Grid[i].Velocity.x = -Grid[i].Velocity.x;
            if(Grid[i].Pos.z>(GRID_SIZE*GRID_HEIGHT) || Grid[i].Pos.z<0)
                Grid[i].Velocity.z = -Grid[i].Velocity.z;

            Grid[i].Velocity.y -= MoveSpeed/2;
            Grid[i].Pos += Grid[i].Velocity*MoveSpeed;
            Changed = TRUE;

            Grid[i].Yaw += Grid[i].YawV*RotateSpeed*0.07;
            Grid[i].Pitch += Grid[i].PitchV*RotateSpeed*0.07;
            Grid[i].Roll += Grid[i].RollV*RotateSpeed*0.07;
        }
        else
        {
            Grid[i].Velocity *= 0.0;
            Grid[i].Active = FALSE;
        }
        Grid[i].Update();
      }
    }
    if(temp)
        Changed = FALSE;
}

void Terra::Draw(void)
{
    Device->SetIndices(IndexBuffer,0);
    Device->SetStreamSource(0,VertexBuffer,sizeof(Vertex));
    Device->SetVertexShader(Vertex_FVF);
    Device->SetTexture(0,Texture);

    long l=0;
    DWORD num_vertices = 8*(GRID_WIDTH);
    DWORD num_to_draw=0; //per loop
    DWORD num_drawn=0; //total
    DWORD num_triangles = 0;

    BYTE *buffer=NULL;
    Vertex verts[8];

    for(long y=0;y<DrawHeight;y++)
    {
        VertexBuffer->Lock(0,0,(BYTE**)&buffer,D3DLOCK_DISCARD);

        for(long x=0;x<DrawWidth;x++)
        {
          l = ((y*GRID_HEIGHT)+x);
          Grid[l].GetVertices(verts);
          memcpy(buffer,verts,sizeof(Grid[l].Vertices));
          buffer+=sizeof(Grid[l].Vertices);
          num_triangles += 6;
        }

        VertexBuffer->Unlock();

        for(;;) //send as many triangles as possible, or send them as much at a time as possible
        {
          if(num_drawn<num_triangles)
          {
            num_to_draw = (MaxPrimitives < (num_triangles-num_drawn)) ? MaxPrimitives : (num_triangles-num_drawn);
            Device->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,num_vertices,0,num_to_draw);
            num_drawn += num_to_draw;
            }
          else
            break;
        }
        num_triangles = 0;
        num_drawn = 0;
        num_to_draw = 0;
    }
}

void Terra::Explode(D3DXVECTOR3 pos,float radius,float magnitude) //ka-boom!
{
    float distance;
    float mag_times_dist=0.0;
    D3DXVECTOR3 normalized(0,0,1);

    for(long i=0;i<(GRID_HEIGHT*GRID_WIDTH);i++)
    {
        distance = DISTANCE_NOY(pos,Grid[i].Pos);
        if(distance > radius)
            continue;
        if(Grid[i].Pos.y>16)
            continue;

        D3DXVec3Normalize(&normalized,&(Grid[i].Pos-pos));

        mag_times_dist = magnitude*(1-(distance/radius));
        Grid[i].Velocity += normalized*mag_times_dist;

        Grid[i].RollV += normalized.x*mag_times_dist;
        Grid[i].PitchV += normalized.z*mag_times_dist;
        Grid[i].YawV += mag_times_dist;
        Grid[i].Bounces = 0;
        Grid[i].Active = TRUE;
    }
}

void Terra::SetColors(UCHAR r,UCHAR g,UCHAR b)
{
    for(long i=0;i<(GRID_HEIGHT*GRID_WIDTH);i++)
    {
        Grid[i].R = (UCHAR)r;
        Grid[i].G = (UCHAR)g;
        Grid[i].B = (UCHAR)b;

        for(int ii=0;ii<8;ii++)
        {
            if(ii<4)
                Grid[i].Vertices[ii].color = D3DCOLOR_RGBA( Grid[i].R, Grid[i].G, Grid[i].B,255);
            else
                Grid[i].Vertices[4].color = D3DCOLOR_RGBA(255,0,0,255);
        }

    }
    Changed = TRUE;
    Grid[0].Active = TRUE;
}

void Terra::SetDefaultColors(void)
{
    float temp;
    for(long i=0;i<(GRID_HEIGHT*GRID_WIDTH);i++)
    {
        temp = RandomNumber(0.5)+0.5;
        Grid[i].R = (UCHAR)(temp*255);
        Grid[i].G = (UCHAR)((RandomNumber(0.25)+0.75)*255);
        Grid[i].B = (UCHAR)(temp*255);

        for(int ii=0;ii<8;ii++)
        {
            if(ii<4)
            Grid[i].Vertices[ii].color = D3DCOLOR_RGBA( Grid[i].R, Grid[i].G, Grid[i].B,255);
            else
                Grid[i].Vertices[4].color = D3DCOLOR_RGBA(255,0,0,255);
        }
        Grid[i].Vertices[4].color = D3DCOLOR_RGBA(255,0,0,255);
    }
    Changed = TRUE;
    Grid[0].Active = TRUE;
}
//---------------------------------------------------------------------------
