//---------------------------------------------------------------------------
#ifndef TerraH
#define TerraH

#include <d3dx8.h>
#include <cstdlib>
#include <cmath>
#include "Vertex.h"
//---------------------------------------------------------------------------

#define GRID_WIDTH  64
#define GRID_HEIGHT 64
#define GRID_SIZE       16

#define RandomNumber(a) (float)(((float)rand()/RAND_MAX)*a)

inline float DISTANCE(D3DXVECTOR3 a,D3DXVECTOR3 b)
{
    static float dist_squared = 0.0;
    dist_squared = ((b.x-a.x)*(b.x-a.x)+ (b.y-a.y)*(b.y-a.y)+ (b.z-a.z)*(b.z-a.z));
    if(dist_squared!=0)
        return(sqrt(dist_squared));
    else
        return(0.0);
}

inline float DISTANCE_NOY(D3DXVECTOR3 a,D3DXVECTOR3 b)
{
    static float dist_squared = 0.0;
    dist_squared = ((b.x-a.x)*(b.x-a.x)+ (b.z-a.z)*(b.z-a.z));
    if(dist_squared!=0)
        return(sqrt(dist_squared));
    else
        return(0.0);
}

#define UCHAR unsigned char

typedef struct _GridInfo
{
    _GridInfo()
    {
        D3DXMatrixIdentity(&Transform);
        Pos = D3DXVECTOR3(0,0,0);
        Velocity = D3DXVECTOR3(0,0,0);
        Yaw = 0.0;
        Pitch = 0.0;
        Roll = 0.0;
        YawV = 0.0;
        PitchV = 0.0;
        RollV = 0.0;
        for(int i=0;i<8;i++)
            Vertices[i] = Vertex();
        R = 255;
        G = 255;
        B = 255;
        Bounces = 10;
        Active = TRUE;
    }

    void Update(void)
    {
        D3DXMATRIX temp;
        D3DXMatrixIdentity(&Transform);
        D3DXMatrixRotationYawPitchRoll(&temp,Yaw,Pitch,Roll);
        Transform *= temp;
        D3DXMatrixTranslation(&temp,Pos.x,Pos.y,Pos.z);
        Transform *= temp;
    }
    unsigned long GetColor(void)
    {
        return(D3DCOLOR_RGBA(R,G,B,255));
    }

    void Init(_GridInfo a,_GridInfo b,_GridInfo c)
    {
        Vertices[0] = Vertex(Pos-Pos,GetColor(),0.25,0.75);
        Vertices[1] = Vertex(a.Pos-Pos,a.GetColor(),0.75,0.75);
        Vertices[2] = Vertex(b.Pos-Pos,b.GetColor(),0.25,0.25);
        Vertices[3] = Vertex(c.Pos-Pos,c.GetColor(),0.75,0.25);

        Vertices[4] = Vertex(D3DXVECTOR3((GRID_SIZE>>1),-(GRID_SIZE),(GRID_SIZE>>1)),D3DCOLOR_RGBA(255,0,0,255),0.5,1.0);
        Vertices[5] = Vertex(D3DXVECTOR3((GRID_SIZE>>1),-(GRID_SIZE),(GRID_SIZE>>1)),D3DCOLOR_RGBA(255,0,0,255),1.0,0.5);
        Vertices[6] = Vertex(D3DXVECTOR3((GRID_SIZE>>1),-(GRID_SIZE),(GRID_SIZE>>1)),D3DCOLOR_RGBA(255,0,0,255),0.5,0.0);
        Vertices[7] = Vertex(D3DXVECTOR3((GRID_SIZE>>1),-(GRID_SIZE),(GRID_SIZE>>1)),D3DCOLOR_RGBA(255,0,0,255),0.0,0.5);

        for(int i=0;i<8;i++)
        {
            Vertices[i].x -= (GRID_SIZE>>1);
            Vertices[i].y -= -(GRID_SIZE>>1);
            Vertices[i].z -= (GRID_SIZE>>1);
        }

        Update();
    }
    void GetVertices(Vertex *verts)
    {
        D3DXVECTOR3 temp;
        for(int i=0;i<8;i++)
        {
            verts[i] = Vertices[i];
            D3DXVec3TransformCoord(&temp,&D3DXVECTOR3(Vertices[i].x,Vertices[i].y,Vertices[i].z),&Transform);
            verts[i].x = temp.x;
            verts[i].y = temp.y;
            verts[i].z = temp.z;
        }
    }

    D3DXMATRIX Transform;
    D3DXVECTOR3 Pos;
    D3DXVECTOR3 Velocity;
    float Yaw,Pitch,Roll;
    float YawV,PitchV,RollV;
    int Bounces;

    BOOL Active;    //is this active (moving) or stationary

    unsigned char R;
    unsigned char G;
    unsigned char B;

    Vertex Vertices[8]; //the untransformed points that make up this earth particle

}GridInfo;

class Terra //The Earth
{
public:
    Terra();
    ~Terra();

    void Initialize(LPDIRECT3DDEVICE8 Dev,D3DXVECTOR3 *v);
    void Update(float MoveSpeed,float RotateSpeed);
    void Draw(void);

    void Explode(D3DXVECTOR3 pos,float radius,float magnitude);
    void SetColors(UCHAR r,UCHAR g,UCHAR b);
    void SetDefaultColors(void);  

    GridInfo Grid[GRID_WIDTH*GRID_HEIGHT];  //the grid of earth particles

    long RockHits[5]; //sounds when the rocks hit each other
    D3DXVECTOR3 *Viewer; //where the viewer is

    long MaxPrimitives;
    BOOL Changed;

    WORD DrawWidth;
    WORD DrawHeight;

    LPDIRECT3DDEVICE8 Device;
    LPDIRECT3DVERTEXBUFFER8 VertexBuffer;
    LPDIRECT3DINDEXBUFFER8 IndexBuffer;
    LPDIRECT3DTEXTURE8 Texture;
};

#endif
