//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "TetraVis_source.h"
#include "ResolutionForm.h"
#include <d3dx8.h>
#include <time.h>
#include "Sound.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
    : TForm(Owner)
{
    //  terra-ignis-aqua-zephyr
    Done = FALSE;
    Camera.Pos = D3DXVECTOR3(0,1,0);
    Camera.LookAt = D3DXVECTOR3(0,0,4);
    Camera.YRot = 0.0;

    MoveSpeed = 0.5;
    RotateSpeed = 0.04;

    BackgroundSprite = NULL;
    Daytime = NULL;
    Nighttime = NULL;
    Scaling = D3DXVECTOR2(1.0,1.0);
    IsDay = TRUE;

    ScreenWidth = 640;
    ScreenHeight = 480;

    FullScreen = FALSE;

    LastTime = 0;
    SecondsPerFrame = 0.0;
    FramesPerSecond = 0.0;

    bEarth = TRUE;
    bFire = TRUE;
    bWater = TRUE;
    bWind = TRUE;

    F1Down = FALSE;
    PauseDown = FALSE;
    Paused = FALSE;
    bRender = TRUE;
    ResetMoveSpeed = FALSE;

    for(int i=0;i<4;i++)
        NumberDown[i] = FALSE;
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormCreate(TObject *Sender)
{
    PostMessage(Handle, WM_MAINLOOP, 0, 0);
}
//---------------------------------------------------------------------------
char *TForm1::InitD3D(void)
{
    ResForm->ShowModal();
    ResForm->Hide();

    int Resolution=ResForm->ComboBox1->ItemIndex;
    if(Resolution<0)
        Resolution = 0;

    FullScreen = ResForm->FullScreenCheckBox->Checked;
    IsDay = ResForm->DaytimeCheckbox->Checked;

    // Create the D3D object.
    if( NULL == ( Direct3D = Direct3DCreate8( D3D_SDK_VERSION ) ) )
        return("Direct3D Object Creation Failed.");

    // Get the current desktop display mode, so we can set up a back
    // buffer of the same format
    D3DDISPLAYMODE d3ddm;
    if( FAILED( Direct3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm ) ) )
        return("Could not get adapter display mode.");

    // Set up the structure used to create the D3DDevice.
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = d3ddm.Format;
    d3dpp.BackBufferCount = 1;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
    d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;

    if(FullScreen)
    {
        d3dpp.Windowed = FALSE;
        d3dpp.SwapEffect = D3DSWAPEFFECT_FLIP;
        switch(Resolution)
        {
            default:
            case 0:
            d3dpp.BackBufferWidth = 640;
            d3dpp.BackBufferHeight = 480;
            break;

            case 1:
            d3dpp.BackBufferWidth = 800;
            d3dpp.BackBufferHeight = 600;
            break;

            case 2:
            d3dpp.BackBufferWidth = 1024;
            d3dpp.BackBufferHeight = 768;
            break;
        }
        d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;

        ShowCursor(FALSE);
    }

    // Create the D3DDevice
    if( FAILED( Direct3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, Handle,
                                      D3DCREATE_HARDWARE_VERTEXPROCESSING,
                                      &d3dpp, &Device ) ) )
    {
        if( FAILED( Direct3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, Handle,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &d3dpp, &Device ) ) )
            return("Could not create Device.");
    }

    // Turn on ambient lighting
    Device->SetRenderState( D3DRS_AMBIENT, 0xffffffff );

    Device->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE );
    Device->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ONE );
    Device->SetRenderState(D3DRS_ZENABLE,TRUE);
    Device->SetRenderState(D3DRS_LIGHTING,FALSE);

    Device->SetTextureStageState(0,D3DTSS_MIPFILTER,D3DTEXF_LINEAR);
    Device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTEXF_LINEAR);
    Device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTEXF_LINEAR);
    Device->SetTextureStageState(1,D3DTSS_MIPFILTER,D3DTEXF_LINEAR);
    Device->SetTextureStageState(1,D3DTSS_MAGFILTER,D3DTEXF_LINEAR);
    Device->SetTextureStageState(1,D3DTSS_MINFILTER,D3DTEXF_LINEAR);

    //World Matrix
    Device->SetTransform( D3DTS_WORLD, D3DXMatrixIdentity(&World) );

    ScreenWidth = Form1->Width;
    ScreenHeight = Form1->Height;

    float AspectRatio = (float)ScreenWidth/ScreenHeight;

    //View Matrix
    Device->SetTransform( D3DTS_VIEW, D3DXMatrixLookAtLH(&View,&D3DXVECTOR3(0,10,-10),&D3DXVECTOR3(0,5,1),&D3DXVECTOR3(0,1,0)) );

    //Projection Matrix
    Device->SetTransform( D3DTS_PROJECTION, D3DXMatrixPerspectiveFovLH(&Projection,D3DXToRadian(70),AspectRatio,0.01, 1000.0) );

    D3DXCreateSprite(Device,&BackgroundSprite);
    D3DXCreateTextureFromFile(Device,"Textures\\Night.jpg",&Nighttime);
    D3DXCreateTextureFromFile(Device,"Textures\\Clouds.jpg",&Daytime);

    InitializeDirectSound(Form1->Handle);

    return("No Error!");
}
//---------------------------------------------------------------------------
void TForm1::InitObjects(void)
{
    Earth = new Terra();
    Earth->Initialize(Device,&Camera.Pos);

    Earth->DrawWidth = ResForm->DrawWidthTrackBar->Position;
    Earth->DrawHeight = ResForm->DrawHeightTrackBar->Position;

    Fire = new Ignis();
    Fire->Initialize(Device);

    Water = new Aqua();
    Water->Initialize(Device,&Camera.Pos);

    Water->DrawWidth = ResForm->DrawWidthTrackBar->Position;
    Water->DrawHeight = ResForm->DrawHeightTrackBar->Position;
    Water->WaterDensity = ResForm->RainDensityTrackBar->Position;
    Water->WaterHeight = ResForm->RainHeightTrackBar->Position;

    Wind = new Zephyr();
    if(!Wind->Initialize(Earth,Fire,Water))
        Done = TRUE;

    Wind->Speed = ((float)ResForm->WindSpeedTrackBar->Position/100);
    Wind->WindDirection = ResForm->wind_dir;
}
//---------------------------------------------------------------------------
void TForm1::RefreshObjects(void)
{
    Earth->DrawWidth = ResForm->DrawWidthTrackBar->Position;
    Earth->DrawHeight = ResForm->DrawHeightTrackBar->Position;

    Water->DrawWidth = ResForm->DrawWidthTrackBar->Position;
    Water->DrawHeight = ResForm->DrawHeightTrackBar->Position;
    Water->WaterDensity = ResForm->RainDensityTrackBar->Position;
    Water->WaterHeight = ResForm->RainHeightTrackBar->Position;

    Wind->Speed = ((float)ResForm->WindSpeedTrackBar->Position/100);
    Wind->WindDirection = ResForm->wind_dir;
}
//---------------------------------------------------------------------------
void TForm1::Init(void)
{
    InitD3D();
    InitObjects();
}
//---------------------------------------------------------------------------
void TForm1::DestroyD3D(void)
{
    DestroyDirectSound();

    SAFE_RELEASE(BackgroundSprite);
    SAFE_RELEASE(Daytime);
    SAFE_RELEASE(Nighttime);

    if(Device != NULL)
        Device->Release();
    Device = NULL;

    if(Direct3D != NULL)
        Direct3D->Release();
    Direct3D = NULL;
}
//---------------------------------------------------------------------------
void TForm1::DestroyObjects(void)
{
    delete Earth;
    delete Fire;
    delete Water;
    delete Wind;
}
//---------------------------------------------------------------------------
void TForm1::UpdateObjects(void)
{
    //scale the background appropriately
    Scaling.x = (float)ScreenWidth/512;
    Scaling.y = (float)ScreenHeight/512;

    //update the time
    SecondsPerFrame = (float)(GetTickCount()-LastTime)/1000;
    if(SecondsPerFrame!=0)
        FramesPerSecond = 1/SecondsPerFrame;
    else
        FramesPerSecond = 0.0;
    LastTime = GetTickCount();

    MoveSpeed = MOVE_SPEED*SecondsPerFrame;
    RotateSpeed = ROTATE_SPEED*SecondsPerFrame;

    if(ResetMoveSpeed)
    {
        MoveSpeed = 0.0;
        RotateSpeed = 0.0;
        ResetMoveSpeed = FALSE;
    }

    if(!Paused && bRender)
    {
     if(bEarth)
        Earth->Update(MoveSpeed,RotateSpeed);
     if(bFire)
        Fire->Update(MoveSpeed,Camera.Pos);
     if(bWater)
        Water->Update(MoveSpeed,Earth);
     if(bWind)
        Wind->Update(MoveSpeed);
    }
}
//---------------------------------------------------------------------------
void TForm1::DrawObjects(void)
{
    //turn off z writes so the background stays in the background
    Device->SetRenderState(D3DRS_ZWRITEENABLE,FALSE);
    Device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    BackgroundSprite->Draw((IsDay) ? Daytime : Nighttime,NULL,&Scaling,NULL,0.0,NULL,D3DCOLOR_RGBA(255,255,255,(UCHAR)255));
    Device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    Device->SetRenderState(D3DRS_ZWRITEENABLE,TRUE);

    if(bEarth)
        Earth->Draw();
    if(bFire)
        Fire->Draw(Camera.YRot);
    if(bWater)
        Water->Draw();
}
//---------------------------------------------------------------------------
void __fastcall TForm1::Main(TMessage Msg)
{
    Init();

    try
    {
     while(!Done)
     {
        UpdateObjects();
        
        if(bRender)
        {

         //update the view matrix based on the camera struct
         Device->SetTransform(D3DTS_VIEW,D3DXMatrixLookAtLH(&View,&Camera.Pos,&Camera.LookAt,&D3DXVECTOR3(0,1,0)));

         Device->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0x00000000, 1.0f, 0 );
         if(FAILED(Device->BeginScene()))
         {
            Done = TRUE;
            break;
         }

         DrawObjects();

         if(FAILED(Device->EndScene()))
         {
            Done = TRUE;
            break;
         }

         if(FAILED((Device->Present(NULL,NULL,NULL,NULL))))
         {
            Done = TRUE;
            break;
         }
        }
        GetInput();
        Application->ProcessMessages();
     }
    }
    catch(Exception &excep)
    {
        DestroyObjects();
        DestroyD3D();
        Application->ShowException(&excep);
        Application->Terminate();
    }
    catch(...)
    {
        DestroyObjects();
        DestroyD3D();
        Application->MessageBox("A Fatal Error Occured. At least the computer didn't crash! (Press SpaceBar to continue...).","Error",MB_OK);
        Application->Terminate();
    }

    DestroyObjects();
    DestroyD3D();
    Application->Terminate();
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormClose(TObject *Sender, TCloseAction &Action)
{
    Done = TRUE;
}
//---------------------------------------------------------------------------
void TForm1::GetInput(void)
{
    if(KEY_DOWN(VK_F2))
    {
        ShowCursor(TRUE);
        bRender = FALSE;
        ResForm->ShowModal();
        RefreshObjects();
        ShowCursor(FALSE);
        bRender = TRUE;
        ResetMoveSpeed = TRUE;
    }

    if(KEY_DOWN(VK_SHIFT))
    {
        Camera.Pos.y += MoveSpeed;
        Camera.LookAt.y += MoveSpeed;
    }
    if(KEY_DOWN(VK_CONTROL))
    {
        Camera.Pos.y -= MoveSpeed;
        Camera.LookAt.y -= MoveSpeed;
    }
    if(KEY_DOWN(VK_UP))
    {
        D3DXVECTOR3 temp(MoveSpeed*sin(Camera.YRot),0,MoveSpeed*cos(Camera.YRot));
        Camera.Pos += temp;
        Camera.LookAt += temp;
    }
    if(KEY_DOWN(VK_DOWN))
    {
        D3DXVECTOR3 temp(MoveSpeed*sin(Camera.YRot),0,MoveSpeed*cos(Camera.YRot));
        Camera.Pos -= temp;
        Camera.LookAt -= temp;
    }
    if(KEY_DOWN(VK_LEFT))
    {
        Camera.YRot -= RotateSpeed;
        Camera.LookAt = Camera.Pos;

        Camera.LookAt.x += sin(Camera.YRot);
        Camera.LookAt.z += cos(Camera.YRot);
    }
    if(KEY_DOWN(VK_RIGHT))
    {
        Camera.YRot += RotateSpeed;
        Camera.LookAt = Camera.Pos;

        Camera.LookAt.x += sin(Camera.YRot);
        Camera.LookAt.z += cos(Camera.YRot);
    }
    if(KEY_DOWN('Z'))
    {
        Camera.YRot = 0.0;

        Camera.Pos.x-=MoveSpeed;
        Camera.LookAt = Camera.Pos;
        Camera.LookAt.y -= 1;
        Camera.LookAt.z+=0.025;
    }
    if(KEY_DOWN('X'))
    {
        Camera.YRot = 0.0;

        Camera.Pos.x+=MoveSpeed;
        Camera.LookAt = Camera.Pos;
        Camera.LookAt.y -= 1;
        Camera.LookAt.z+=0.025;
    }

    if(KEY_DOWN(VK_SPACE)) //kaboom!!
    {
        if(bFire)
            Fire->Explode(D3DXVECTOR3(rand()%GRID_WIDTH,-RandomNumber(1024)+32,rand()%GRID_HEIGHT),(long)RandomNumber(16),RandomNumber(25)+5,bEarth ? Earth : NULL);
        else
            Earth->Explode(D3DXVECTOR3(RandomNumber(Earth->DrawWidth*GRID_SIZE),-RandomNumber(1024)+32,RandomNumber(Earth->DrawHeight*GRID_SIZE)),RandomNumber(16*GRID_SIZE),RandomNumber(25)+5);            
    }

    if(KEY_DOWN(VK_F1))
    {
        F1Down = TRUE;
    }
    else
    if(F1Down)
    {
        IsDay = !IsDay;
        if(IsDay)
            Earth->SetDefaultColors();
        else
            Earth->SetColors(0,64,128);
        F1Down = FALSE;
    }

    if(KEY_DOWN(VK_RETURN))
    {
        PauseDown = TRUE;
    }
    else
    if(PauseDown)
    {
        Paused = !Paused;
        PauseDown = FALSE;
    }

    if(KEY_DOWN('1'))
    {
        NumberDown[0] = TRUE;
    }
    else
    if(NumberDown[0])
    {
        bEarth = !bEarth;
        NumberDown[0] = FALSE;
    }

    if(KEY_DOWN('2'))
    {
        NumberDown[1] = TRUE;
    }
    else
    if(NumberDown[1])
    {
        bFire = !bFire;
        NumberDown[1] = FALSE;

        if(bFire)
            PlaySound(Fire->LavaGurgleSound,TRUE);
        else
            StopSound(Fire->LavaGurgleSound);
    }

    if(KEY_DOWN('3'))
    {
        NumberDown[2] = TRUE;
    }
    else
    if(NumberDown[2])
    {
        bWater = !bWater;
        NumberDown[2] = FALSE;

        if(bWater)
            PlaySound(Water->RainSound,TRUE);
        else
            StopSound(Water->RainSound);
    }

    if(KEY_DOWN('4'))
    {
        NumberDown[3] = TRUE;
    }
    else
    if(NumberDown[3])
    {
        bWind = !bWind;
        NumberDown[3] = FALSE;

        if(bWind)
            PlaySound(Wind->WindSound,TRUE);
        else
            StopSound(Wind->WindSound);
    }

    if(FullScreen && KEY_DOWN(VK_ESCAPE))
    {
        Done = TRUE;
    }
}
//---------------------------------------------------------------------------
