//---------------------------------------------------------------------------
#ifndef TetraVis_sourceH
#define TetraVis_sourceH
//---------------------------------------------------------------------------
#include <Classes.hpp>
#include <Controls.hpp>
#include <StdCtrls.hpp>
#include <Forms.hpp>
//---------------------------------------------------------------------------
#include "Vertex.h"
#include <d3dx8.h>
#include "Ignis.h"
#include "Terra.h"
#include "Aqua.h"
#include "Zephyr.h"

#define WM_MAINLOOP WM_USER
#define KEY_DOWN(vk_code) ((GetAsyncKeyState(vk_code) & 0x8000) ? 1 : 0)

//speed per second
#define MOVE_SPEED      28
#define ROTATE_SPEED    D3DX_PI/2

class TForm1 : public TForm
{
__published:	// IDE-managed Components
    void __fastcall FormCreate(TObject *Sender);
    void __fastcall FormClose(TObject *Sender, TCloseAction &Action);
private:	// User declarations
public:		// User declarations

    //initialization functions
    void Init(void);
    char *InitD3D(void);
    void InitObjects(void);
    void RefreshObjects(void);

    //shut-down functions
    void DestroyObjects(void);
    void DestroyD3D(void);

    //main-loop
    void GetInput(void);
    void UpdateObjects(void);
    void DrawObjects(void);
    void __fastcall Main(TMessage Msg);

    BEGIN_MESSAGE_MAP
        MESSAGE_HANDLER(WM_MAINLOOP, TMessage, Main)
    END_MESSAGE_MAP(TForm)


    //D3D Variables
    LPDIRECT3D8 Direct3D;
    LPDIRECT3DDEVICE8 Device;
    DWORD VertexShader;

    D3DXMATRIX World,View,Projection;

    //Program variables
    int ScreenWidth;
    int ScreenHeight;
    BOOL FullScreen;
    BOOL Done;
    struct _Camera
    {
        D3DXVECTOR3 Pos;
        D3DXVECTOR3 LookAt;

        float YRot;

    }Camera;
    float MoveSpeed;    //the speed the camera moves (units per frame)
    float RotateSpeed;  //the speed the camera rotates (radians per frame)

    //this is used to draw the background
    D3DXVECTOR2 Scaling;
    LPD3DXSPRITE BackgroundSprite;
    LPDIRECT3DTEXTURE8 Daytime;
    LPDIRECT3DTEXTURE8 Nighttime;
    BOOL IsDay;
    BOOL F1Down;

    //timing variables
    DWORD LastTime;
    float SecondsPerFrame;
    float FramesPerSecond;

    //input helper variables
    BOOL NumberDown[4]; //describes a number's state on the keyboard, used to count up-down clicks
    BOOL bRender;
    BOOL Paused;
    BOOL PauseDown;
    BOOL ResetMoveSpeed;

    //The Four Elements
    Terra *Earth;
    Ignis *Fire;
    Aqua *Water;
    Zephyr *Wind;

    BOOL bEarth; //draw them, or not?
    BOOL bFire;
    BOOL bWater;
    BOOL bWind;

    __fastcall TForm1(TComponent* Owner);
};
//---------------------------------------------------------------------------
extern PACKAGE TForm1 *Form1;
//---------------------------------------------------------------------------
#endif
