//---------------------------------------------------------------------------
#ifndef VertexH
#define VertexH

#include <d3dx8.h>
//---------------------------------------------------------------------------

#define SAFE_RELEASE(a) if(a!=NULL) \
                        {\
                            a->Release();\
                            a=NULL;\
                        }

#define Vertex_FVF (D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0))

typedef struct _Vertex  //basic vertex structure with 4 useful constructors
{
    _Vertex()
    {
        x = 0.0;
        y = 0.0;
        z = 0.0;
        color = D3DCOLOR_ARGB(255,255,255,255);
        u = 0.0;
        v = 0.0;
    }
    _Vertex(D3DXVECTOR3 vec)
    {
        x = vec.x;
        y = vec.y;
        z = vec.z;
        color = D3DCOLOR_ARGB(255,255,255,255);
        u = 0.0;
        v = 0.0;
    }
    _Vertex(D3DXVECTOR3 vec,DWORD _color)
    {
        x = vec.x;
        y = vec.y;
        z = vec.z;
        color = _color;
        u = 0.0;
        v = 0.0;
    }
    _Vertex(D3DXVECTOR3 vec,DWORD _color,float tu,float tv)
    {
        x = vec.x;
        y = vec.y;
        z = vec.z;
        color = _color;
        u = tu;
        v = tv;
    }

    D3DXVECTOR3 ToVector(void)
    {
        return(D3DXVECTOR3(x,y,z));
    };

    float x,y,z;
    DWORD color;
    float u,v;

}Vertex;

#define ParticleVertex_FVF (D3DFVF_XYZ | D3DFVF_DIFFUSE)
inline DWORD FToDW(float a) {return *((DWORD*)(&a));}

typedef struct _ParticleVertex  //basic particle vertex structure with 4 useful constructors
{
    _ParticleVertex()
    {
        x = 0.0;
        y = 0.0;
        z = 0.0;
        color = D3DCOLOR_ARGB(255,255,255,255);
    }
    _ParticleVertex(D3DXVECTOR3 vec)
    {
        x = vec.x;
        y = vec.y;
        z = vec.z;
        color = D3DCOLOR_ARGB(255,255,255,255);
    }
    _ParticleVertex(D3DXVECTOR3 vec,DWORD _color)
    {
        x = vec.x;
        y = vec.y;
        z = vec.z;
        color = _color;
    }
    _ParticleVertex(D3DXVECTOR3 vec,DWORD _color,float _size)
    {
        x = vec.x;
        y = vec.y;
        z = vec.z;
        color = _color;
    }

    float x;
    float y;
    float z;
    DWORD color;

}ParticleVertex;

#endif
